/****************************************************************************
    Poti.c
    ------
    bungsprojekt zum Einlesen der Stellung eines Potentiometers.
    Die Stellung des Poti in Prozent (0...100) soll jeweils angezeigt werden.
    Die Messung erfolgt dadurch, dass ein Kondensator mit dem Poti in Reihe
    geschaltet wird. Im Knoten dazwischen fhren wird die Leitung ber einen
    Schutzwiderstand von etwa 220 Ohm an den Messpin (Port D, Pin 2).
    An Pin 3 von Port D wird ein Taster gegen Masse angeschlossen. Wird
    dieser Taster gedrckt (gedrckt halten) kann das Poti kalibriert werden,
    indem mehrfach vall nach links und nach rechts gedreht wird. Die
    minimale und maximale Wert wird daraufhin im EEPROM gespeichert.

    Autor:    Christian Schifferle
    Datum:    November 2002
    Software: AVR-GCC
    needed
****************************************************************************/

#include <io.h>	// Wir brauchen Zugriff auf die I/O's des Controllers
#include <progmem.h>
#include <eeprom.h>

#include "..\Allgemein\UartPrintF.h"

// Typendefinition fr ein BYTE
typedef unsigned char  BYTE;
typedef unsigned int   WORD;

#include "Pot.h"


#define F_CPU            8000000      /* 8Mhz */
#define UART_BAUD_RATE      9600      /* 9600 baud */

// Berechnung des Inhalts des UBBR-Registers
#define UART_BAUD_SELECT (F_CPU/(UART_BAUD_RATE*16L)-1)

#define CALIBRATE

// Funktionsprototypen
#ifdef CALIBRATE
void Calibrate (void);
#endif

// Kalibrierungsdaten des Poti. Werden zustzlich im EEPROM gespeichert.
struct {
    WORD nMin;  // Messwert bei Minimalstellung
    WORD nMax;  // Messwert bei Maximalstellung
} kalib;

// Hilfsvariablen
BYTE *pb;
WORD n;
WORD nLastPoti;	// Letzter gelesener Poti-Wert

// Globaler Puffer zur Datenbernahme aus dem Programmspeicher
char szPmBuf[42];

int main (void)
/****************************************************************************
    Hier ist der Programmeinstiegspunkt.
*/
{
    // Die zu sendenden Strings werden im Programmspeicher abgelegt, weil es
    // im SRAM zu wenig Platz hat.
    // Bei Bedarf werden die Strings in die globale Variable szPmBuf kopiert.
    // Man erkennt dies daran, dass die entsprechenden Strings im ROM-File
    // ziemlich am Anfang zu finden sind. Normal deklarierte Strings, welche
    // im SRAM abgelegt werden, befinden sich im ROM-File am Ende.
    // Dies ist brigens die einzige Art, Strings wirklich im Programmspeicher
    // (Flash) abzulegen. Alle anderen Methoden wie prog_char etc. haben
    // irgendwo einen Hund drin.
    char *ps1 = PSTR("Messung %5u: %3u (%5u)\r");
    char *ps2 = PSTR("Zu hoch\r");
    register WORD k, kOld;

    /* UART Control Register laden. Wir verwenden (noch)
       keine Interrupts. Vorerst wollen wir nur mal senden, also wird
       nur das Transmit Enable Flag gesetzt. */
    outp ((1 << TXEN), UCR);

    /* Baudrate einstellen */
    outp ((BYTE) UART_BAUD_SELECT, UBRR);          

    /* Hier beginnt die Endlosschleife. Dies knnte auch als
       while (1)
       definiert werden. Der Ausdruck for (;;) ist jedoch in C
       eine bliche Schreibweise fr sowas.
    */

    // Pin 4 als Ausgang schalten als Triggersignal
    sbi (DDRD, PIND4);

#ifdef CALIBRATE
    // Kalibrierung aus dem EEPROM auslesen
    eeprom_read_block (&kalib, 0, sizeof kalib);

    outp (0, DDRD);       // Port D als Eingang schalten.
    outp (0x08, PORTD);   // Pull-Up auf Pin 3 aktivieren.
                          // Mit dem Taster an Pin3 wird die Kalibrierung
                          // eingeleitet.
#endif

Loop:
#ifdef CALIBRATE
    if (bit_is_clear (PIND, PIND3)) {
        Calibrate ();
        n = 0;
    }
#endif
    // Triggern
    sbi (PORTD, PIND4);
    nLastPoti = Poti (DDRD, PORTD, PIND, PIND2);
    cbi (PORTD, PIND4);
    // Wert anhand der Kalibrationsdaten in einen Bereich
    // von 0 bis 100 umwandeln (Strahlensatz) (x = nLastPoti).
    //
    // kalib.nMax |                        /
    //            |                      /
    //            |                    /
    //          x |----------------- x
    //            |                / |
    //            |              /   |
    //            |            /     |     k          100
    //            |          /       |   -------- = -------
    //            |        /         |   x - nMin   Max - Min
    //            |      /           |
    //            |    /             |              100 (x - Min)
    //            |  /               |          k = -------------
    // kalib.nMin |/                 |                Max - Min
    //            |                  |
    //            |                  |
    //          0 +---------------------------
    //            0                  k     100
    //
    // Achtung! Die Berechnung muss mit LONG-Variablen erfolgen, da es
    // sonst einen Integerberlauf gibt.
    if (nLastPoti < 65535) {
        k = 100L * (long) (nLastPoti - kalib.nMin)
            / (long) (kalib.nMax - kalib.nMin);
        n++;
        /* Um das SRAM zu entlasten werden die Formatstrings im
           Programmspeicher abgelegt. Hier wird jetzt der String aus dem
           Programmspeicher in die Puffervariable szPmBuf, welche sich im
           SRAM befindet, kopiert, */
	if (k != kOld) {
            strcpy_P (szPmBuf, ps1);
            UartPrintF (szPmBuf, n, k, nLastPoti);
            kOld = k;
        }
    }
    else {
        strcpy_P (szPmBuf, ps2);
        UartPrintF (szPmBuf);
    }
    goto Loop;

} /* end of main */



#ifdef CALIBRATE
void Calibrate (void)
/***********************************************************************
    Kalibriert das Poti. Bei Aufruf dieser Funktion sollte das Poti
    in der Mittelstellung stehen.
    Die Kalibrierung wird in's EEPROM geschrieben.
*/
{
    char *ps2 = PSTR("Min: %5u, Max: %5u \r");

    kalib.nMin = kalib.nMax = nLastPoti;
    UartPrintF ("\n");

    while (bit_is_clear (PIND, PIND3)) {
        sbi (PORTD, PIND4);
        n = Poti (DDRD, PORTD, PIND, PIND2);
        cbi (PORTD, PIND4);
        if (n && n < kalib.nMin) kalib.nMin = n;
        if (n < 65535 && n > kalib.nMax) kalib.nMax = n;
        strcpy_P (szPmBuf, ps2);
        UartPrintF (szPmBuf, kalib.nMin, kalib.nMax);
    }

    // Kalibrierung in's EEPROM schreiben
    n = 0;
    pb = (BYTE *) &kalib;
    for (n=0; n<sizeof kalib; n++, pb++) {
        while (!eeprom_is_ready ()) /* warten, bis EEPROM bereit */;
        eeprom_wb (n, *pb);         // nchstes Byte schreiben
    }

} /* end of Calibrate */
#endif
